@@ -94,6 +94,23 @@ void calc_mean_and_sd_t(size_t ix_arr[], size_t st, size_t end, real_t_ *restric
     }
 }
 
+template <class real_t_>
+double calc_mean_only(size_t ix_arr[], size_t st, size_t end, real_t_ *restrict x)
+{
+    size_t cnt = 0;
+    double m = 0;
+    for (size_t row = st; row <= end; row++)
+    {
+        if (!is_na_or_inf(x[ix_arr[row]]))
+        {
+            cnt++;
+            m += (x[ix_arr[row]] - m) / (double)cnt;
+        }
+    }
+
+    return m;
+}
+
 template <class real_t_>
 void calc_mean_and_sd(size_t ix_arr[], size_t st, size_t end, real_t_ *restrict x,
                       MissingAction missing_action, double &x_sd, double &x_mean)
@@ -135,6 +152,25 @@ void calc_mean_and_sd_weighted(size_t ix_arr[], size_t st, size_t end, real_t_ *
     x_sd   = std::sqrt(s / (long double)cnt);
 }
 
+template <class real_t_, class mapping>
+double calc_mean_only_weighted(size_t ix_arr[], size_t st, size_t end, real_t_ *restrict x, mapping w)
+{
+    double cnt = 0;
+    double w_this;
+    double m = 0;
+    for (size_t row = st; row <= end; row++)
+    {
+        if (!is_na_or_inf(x[ix_arr[row]]))
+        {
+            w_this = w[ix_arr[row]];
+            cnt += w_this;
+            m += w_this * (x[ix_arr[row]] - m) / cnt;
+        }
+    }
+
+    return m;
+}
+
 /* for sparse numerical */
 template <class real_t_, class sparse_ix, class real_t>
 void calc_mean_and_sd(size_t ix_arr[], size_t st, size_t end, size_t col_num,
@@ -227,6 +263,56 @@ void calc_mean_and_sd(size_t ix_arr[], size_t st, size_t end, size_t col_num,
     x_sd = std::fmax(SD_MIN, x_sd);
 }
 
+template <class real_t_, class sparse_ix>
+double calc_mean_only(size_t ix_arr[], size_t st, size_t end, size_t col_num,
+                      real_t_ *restrict Xc, sparse_ix Xc_ind[], sparse_ix Xc_indptr[])
+{
+    /* ix_arr must be already sorted beforehand */
+    if (Xc_indptr[col_num] == Xc_indptr[col_num + 1])
+        return 0;
+    size_t st_col  = Xc_indptr[col_num];
+    size_t end_col = Xc_indptr[col_num + 1] - 1;
+    size_t curr_pos = st_col;
+    size_t ind_end_col = (size_t) Xc_ind[end_col];
+    size_t *ptr_st = std::lower_bound(ix_arr + st, ix_arr + end + 1, (size_t)Xc_ind[st_col]);
+
+    size_t cnt = end - st + 1;
+    size_t added = 0;
+    double m = 0;
+
+    for (size_t *row = ptr_st;
+         row != ix_arr + end + 1 && curr_pos != end_col + 1 && ind_end_col >= *row;
+        )
+    {
+        if (Xc_ind[curr_pos] == (sparse_ix)(*row))
+        {
+            if (is_na_or_inf(Xc[curr_pos]))
+                cnt--;
+            else
+                m += (Xc[curr_pos] - m) / (double)(++added);
+
+            if (row == ix_arr + end || curr_pos == end_col) break;
+            curr_pos = std::lower_bound(Xc_ind + curr_pos + 1, Xc_ind + end_col + 1, *(++row)) - Xc_ind;
+        }
+
+        else
+        {
+            if (Xc_ind[curr_pos] > (sparse_ix)(*row))
+                row = std::lower_bound(row + 1, ix_arr + end + 1, Xc_ind[curr_pos]);
+            else
+                curr_pos = std::lower_bound(Xc_ind + curr_pos + 1, Xc_ind + end_col + 1, *row) - Xc_ind;
+        }
+    }
+
+    if (added == 0)
+        return 0;
+
+    if (cnt > added)
+        m *= ((long double)added / (long double)cnt);
+
+    return m;
+}
+
 template <class real_t_, class sparse_ix, class mapping>
 void calc_mean_and_sd_weighted(size_t ix_arr[], size_t st, size_t end, size_t col_num,
                                real_t_ *restrict Xc, sparse_ix Xc_ind[], sparse_ix Xc_indptr[],
@@ -311,6 +397,65 @@ void calc_mean_and_sd_weighted(size_t ix_arr[], size_t st, size_t end, size_t co
     x_sd   = std::sqrt(s / cnt);
 }
 
+template <class real_t_, class sparse_ix, class mapping>
+double calc_mean_only_weighted(size_t ix_arr[], size_t st, size_t end, size_t col_num,
+                               real_t_ *restrict Xc, sparse_ix Xc_ind[], sparse_ix Xc_indptr[],
+                               mapping w)
+{
+    /* ix_arr must be already sorted beforehand */
+    if (Xc_indptr[col_num] == Xc_indptr[col_num + 1])
+        return 0;
+    size_t st_col  = Xc_indptr[col_num];
+    size_t end_col = Xc_indptr[col_num + 1] - 1;
+    size_t curr_pos = st_col;
+    size_t ind_end_col = (size_t) Xc_ind[end_col];
+    size_t *ptr_st = std::lower_bound(ix_arr + st, ix_arr + end + 1, (size_t)Xc_ind[st_col]);
+
+    long double cnt = 0.;
+    for (size_t row = st; row <= end; row++)
+        cnt += w[ix_arr[row]];
+    long double added = 0;
+    double m = 0;
+    double w_this;
+
+    for (size_t *row = ptr_st;
+         row != ix_arr + end + 1 && curr_pos != end_col + 1 && ind_end_col >= *row;
+        )
+    {
+        if (Xc_ind[curr_pos] == (sparse_ix)(*row))
+        {
+            if (is_na_or_inf(Xc[curr_pos])) {
+                cnt -= w[*row];
+            }
+
+            else {
+                w_this = w[*row];
+                added += w_this;
+                m += w_this * (Xc[curr_pos] - m) / added;
+            }
+
+            if (row == ix_arr + end || curr_pos == end_col) break;
+            curr_pos = std::lower_bound(Xc_ind + curr_pos + 1, Xc_ind + end_col + 1, *(++row)) - Xc_ind;
+        }
+
+        else
+        {
+            if (Xc_ind[curr_pos] > (sparse_ix)(*row))
+                row = std::lower_bound(row + 1, ix_arr + end + 1, Xc_ind[curr_pos]);
+            else
+                curr_pos = std::lower_bound(Xc_ind + curr_pos + 1, Xc_ind + end_col + 1, *row) - Xc_ind;
+        }
+    }
+
+    if (added == 0)
+        return 0;
+
+    if (cnt > added)
+        m *= added / cnt;
+
+    return m;
+}
+
 /* Note about these functions: they write into an array that does not need to match to 'ix_arr',
    and instead, the index that is stored in ix_arr[n] will have the value in res[n] */
 
